"""
Kakao OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/kakao.html
"""
from .oauth import BaseOAuth2


class KakaoOAuth2(BaseOAuth2):
    """Kakao OAuth authentication backend"""
    name = 'kakao'
    AUTHORIZATION_URL = 'https://kauth.kakao.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://kauth.kakao.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def get_user_id(self, details, response):
        return response['id']

    def get_user_details(self, response):
        """Return user details from Kakao account"""
        nickname = response['properties']['nickname']
        return {
            'username': nickname,
            'email': '',
            'fullname': '',
            'first_name': '',
            'last_name': ''
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('https://kapi.kakao.com/v1/user/me',
                             params={'access_token': access_token})

    def auth_complete_params(self, state=None):
        return {
            'grant_type': 'authorization_code',
            'code': self.data.get('code', ''),
            'client_id': self.get_key_and_secret()[0],
        }



import json

from six.moves.urllib_parse import urlencode


from .oauth import OAuth1Test


class KhanAcademyOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.khanacademy.KhanAcademyOAuth1'
    user_data_url = 'https://www.khanacademy.org/api/v1/user'
    expected_username = 'foo@bar.com'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    user_data_body = json.dumps({
        "email": "foo@bar.com",
        "user_id": "http://googleid.khanacademy.org/11111111111111",
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class MixcloudOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.mixcloud.MixcloudOAuth2'
    user_data_url = 'https://api.mixcloud.com/me/'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'username': 'foobar',
        'cloudcast_count': 0,
        'following_count': 0,
        'url': 'http://www.mixcloud.com/foobar/',
        'pictures': {
            'medium': 'http://images-mix.netdna-ssl.com/w/100/h/100/q/85/'
                      'images/graphics/33_Profile/default_user_600x600-v4.png',
            '320wx320h': 'http://images-mix.netdna-ssl.com/w/320/h/320/q/85/'
                         'images/graphics/33_Profile/'
                         'default_user_600x600-v4.png',
            'extra_large': 'http://images-mix.netdna-ssl.com/w/600/h/600/q/85/'
                           'images/graphics/33_Profile/'
                           'default_user_600x600-v4.png',
            'large': 'http://images-mix.netdna-ssl.com/w/300/h/300/q/85/'
                     'images/graphics/33_Profile/default_user_600x600-v4.png',
            '640wx640h': 'http://images-mix.netdna-ssl.com/w/640/h/640/q/85/'
                         'images/graphics/33_Profile/'
                         'default_user_600x600-v4.png',
            'medium_mobile': 'http://images-mix.netdna-ssl.com/w/80/h/80/q/75/'
                             'images/graphics/33_Profile/'
                             'default_user_600x600-v4.png',
            'small': 'http://images-mix.netdna-ssl.com/w/25/h/25/q/85/images/'
                     'graphics/33_Profile/default_user_600x600-v4.png',
            'thumbnail': 'http://images-mix.netdna-ssl.com/w/50/h/50/q/85/'
                         'images/graphics/33_Profile/'
                         'default_user_600x600-v4.png'
        },
        'is_current_user': True,
        'listen_count': 0,
        'updated_time': '2013-03-17T06:26:31Z',
        'following': False,
        'follower': False,
        'key': '/foobar/',
        'created_time': '2013-03-17T06:26:31Z',
        'follower_count': 0,
        'favorite_count': 0,
        'name': 'foobar'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class OrbiOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.orbi.OrbiOAuth2'
    user_data_url = 'https://login.orbi.kr/oauth/user/get'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
    })
    user_data_body = json.dumps({
        'username': 'foobar',
        'first_name': 'Foo',
        'last_name': 'Bar',
        'name': 'Foo Bar',

        'imin': '100000',
        'nick': 'foobar',
        'photo': 'http://s3.orbi.kr/data/member/wi/wizetdev_132894975780.jpeg',
        'sex': 'M',
        'birth': '1973-08-03',
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Moves OAuth2 backend, docs at:
    https://dev.moves-app.com/docs/authentication

Written by Avi Alkalay <avi at unix dot sh>
Certified to work with Django 1.6
"""
from .oauth import BaseOAuth2


class MovesOAuth2(BaseOAuth2):
    """Moves OAuth authentication backend"""
    name = 'moves'
    ID_KEY = 'user_id'
    AUTHORIZATION_URL = 'https://api.moves-app.com/oauth/v1/authorize'
    ACCESS_TOKEN_URL = 'https://api.moves-app.com/oauth/v1/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [
        ('refresh_token', 'refresh_token', True),
        ('expires_in', 'expires'),
    ]

    def get_user_details(self, response):
        """Return user details Moves account"""
        return {'username': str(response.get('user_id'))}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('https://api.moves-app.com/api/1.1/user/profile',
                             params={'access_token': access_token})



"""
Facebook OAuth2 and Canvas Application backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/facebook.html
"""
import hmac
import time
import json
import base64
import hashlib

from ..utils import parse_qs, constant_time_compare, handle_http_errors
from .oauth import BaseOAuth2
from ..exceptions import AuthException, AuthCanceled, AuthUnknownError, \
                         AuthMissingParameter


API_VERSION = 2.8


class FacebookOAuth2(BaseOAuth2):
    """Facebook OAuth2 authentication backend"""
    name = 'facebook'
    RESPONSE_TYPE = None
    SCOPE_SEPARATOR = ','
    AUTHORIZATION_URL = 'https://www.facebook.com/v{version}/dialog/oauth'
    ACCESS_TOKEN_URL = \
        'https://graph.facebook.com/v{version}/oauth/access_token'
    REVOKE_TOKEN_URL = \
        'https://graph.facebook.com/v{version}/{uid}/permissions'
    REVOKE_TOKEN_METHOD = 'DELETE'
    USER_DATA_URL = 'https://graph.facebook.com/v{version}/me'
    EXTRA_DATA = [
        ('id', 'id'),
        ('expires', 'expires'),
        ('granted_scopes', 'granted_scopes'),
        ('denied_scopes', 'denied_scopes')
    ]

    def auth_params(self, state=None):
        params = super(FacebookOAuth2, self).auth_params(state)
        params['return_scopes'] = 'true'
        return params

    def authorization_url(self):
        version = self.setting('API_VERSION', API_VERSION)
        return self.AUTHORIZATION_URL.format(version=version)

    def access_token_url(self):
        version = self.setting('API_VERSION', API_VERSION)
        return self.ACCESS_TOKEN_URL.format(version=version)

    def get_user_details(self, response):
        """Return user details from Facebook account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('name', ''),
            response.get('first_name', ''),
            response.get('last_name', '')
        )
        return {'username': response.get('username', response.get('name')),
                'email': response.get('email', ''),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        params = self.setting('PROFILE_EXTRA_PARAMS', {})
        params['access_token'] = access_token

        if self.setting('APPSECRET_PROOF', True):
            _, secret = self.get_key_and_secret()
            params['appsecret_proof'] = hmac.new(
                secret.encode('utf8'),
                msg=access_token.encode('utf8'),
                digestmod=hashlib.sha256
            ).hexdigest()

        version = self.setting('API_VERSION', API_VERSION)
        return self.get_json(self.USER_DATA_URL.format(version=version),
                             params=params)

    def process_error(self, data):
        super(FacebookOAuth2, self).process_error(data)
        if data.get('error_code'):
            raise AuthCanceled(self, data.get('error_message') or
                                     data.get('error_code'))

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes loging process, must return user instance"""
        self.process_error(self.data)
        if not self.data.get('code'):
            raise AuthMissingParameter(self, 'code')
        state = self.validate_state()
        key, secret = self.get_key_and_secret()
        response = self.request(self.access_token_url(), params={
            'client_id': key,
            'redirect_uri': self.get_redirect_uri(state),
            'client_secret': secret,
            'code': self.data['code']
        })
        # API v2.3 returns a JSON, according to the documents linked at issue
        # #592, but it seems that this needs to be enabled(?), otherwise the
        # usual querystring type response is returned.
        try:
            response = response.json()
        except ValueError:
            response = parse_qs(response.text)
        access_token = response['access_token']
        return self.do_auth(access_token, response, *args, **kwargs)

    def process_refresh_token_response(self, response, *args, **kwargs):
        try:
            return response.json()
        except ValueError:
            return parse_qs(response.content)

    def refresh_token_params(self, token, *args, **kwargs):
        client_id, client_secret = self.get_key_and_secret()
        return {
            'fb_exchange_token': token,
            'grant_type': 'fb_exchange_token',
            'client_id': client_id,
            'client_secret': client_secret
        }

    def do_auth(self, access_token, response=None, *args, **kwargs):
        response = response or {}

        data = self.user_data(access_token)

        if not isinstance(data, dict):
            # From time to time Facebook responds back a JSON with just
            # False as value, the reason is still unknown, but since the
            # data is needed (it contains the user ID used to identify the
            # account on further logins), this app cannot allow it to
            # continue with the auth process.
            raise AuthUnknownError(self, 'An error ocurred while retrieving '
                                         'users Facebook data')

        data['access_token'] = access_token
        if 'expires_in' in response:
            data['expires'] = response['expires_in']

        if self.data.get('granted_scopes'):
            data['granted_scopes'] = self.data['granted_scopes'].split(',')

        if self.data.get('denied_scopes'):
            data['denied_scopes'] = self.data['denied_scopes'].split(',')

        kwargs.update({'backend': self, 'response': data})
        return self.strategy.authenticate(*args, **kwargs)

    def revoke_token_url(self, token, uid):
        version = self.setting('API_VERSION', API_VERSION)
        return self.REVOKE_TOKEN_URL.format(version=version, uid=uid)

    def revoke_token_params(self, token, uid):
        return {'access_token': token}

    def process_revoke_token_response(self, response):
        return super(FacebookOAuth2, self).process_revoke_token_response(
            response
        ) and response.content == 'true'


class FacebookAppOAuth2(FacebookOAuth2):
    """Facebook Application Authentication support"""
    name = 'facebook-app'

    def uses_redirect(self):
        return False

    def auth_complete(self, *args, **kwargs):
        access_token = None
        response = {}

        if 'signed_request' in self.data:
            key, secret = self.get_key_and_secret()
            response = self.load_signed_request(self.data['signed_request'])
            if 'user_id' not in response and 'oauth_token' not in response:
                raise AuthException(self)

            if response is not None:
                access_token = response.get('access_token') or \
                               response.get('oauth_token') or \
                               self.data.get('access_token')

        if access_token is None:
            if self.data.get('error') == 'access_denied':
                raise AuthCanceled(self)
            else:
                raise AuthException(self)
        return self.do_auth(access_token, response, *args, **kwargs)

    def auth_html(self):
        key, secret = self.get_key_and_secret()
        namespace = self.setting('NAMESPACE', None)
        scope = self.setting('SCOPE', '')
        if scope:
            scope = self.SCOPE_SEPARATOR.join(scope)
        ctx = {
            'FACEBOOK_APP_NAMESPACE': namespace or key,
            'FACEBOOK_KEY': key,
            'FACEBOOK_EXTENDED_PERMISSIONS': scope,
            'FACEBOOK_COMPLETE_URI': self.redirect_uri,
        }
        tpl = self.setting('LOCAL_HTML', 'facebook.html')
        return self.strategy.render_html(tpl=tpl, context=ctx)

    def load_signed_request(self, signed_request):
        def base64_url_decode(data):
            data = data.encode('ascii')
            data += '='.encode('ascii') * (4 - (len(data) % 4))
            return base64.urlsafe_b64decode(data)

        key, secret = self.get_key_and_secret()
        try:
            sig, payload = signed_request.split('.', 1)
        except ValueError:
            pass  # ignore if can't split on dot
        else:
            sig = base64_url_decode(sig)
            payload_json_bytes = base64_url_decode(payload)
            data = json.loads(payload_json_bytes.decode('utf-8', 'replace'))
            expected_sig = hmac.new(secret.encode('ascii'),
                                    msg=payload.encode('ascii'),
                                    digestmod=hashlib.sha256).digest()
            # allow the signed_request to function for upto 1 day
            if constant_time_compare(sig, expected_sig) and \
               data['issued_at'] > (time.time() - 86400):
                return data



import re
import sys
import time
import unicodedata
import collections
import functools
import logging

import six
import requests
import social_core

from six.moves.urllib_parse import urlparse, urlunparse, urlencode, \
                                   parse_qs as battery_parse_qs

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.poolmanager import PoolManager

from .exceptions import AuthCanceled, AuthUnreachableProvider


SETTING_PREFIX = 'SOCIAL_AUTH'

social_logger = logging.getLogger('social')


class SSLHttpAdapter(HTTPAdapter):
    """"
    Transport adapter that allows to use any SSL protocol. Based on:
    http://requests.rtfd.org/latest/user/advanced/#example-specific-ssl-version
    """
    def __init__(self, ssl_protocol):
        self.ssl_protocol = ssl_protocol
        super(SSLHttpAdapter, self).__init__()

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            ssl_version=self.ssl_protocol
        )

    @classmethod
    def ssl_adapter_session(cls, ssl_protocol):
        session = requests.Session()
        session.mount('https://', SSLHttpAdapter(ssl_protocol))
        return session


def import_module(name):
    __import__(name)
    return sys.modules[name]


def module_member(name):
    mod, member = name.rsplit('.', 1)
    module = import_module(mod)
    return getattr(module, member)


def user_agent():
    """Builds a simple User-Agent string to send in requests"""
    return 'social-auth-' + social_core.__version__


def url_add_parameters(url, params):
    """Adds parameters to URL, parameter will be repeated if already present"""
    if params:
        fragments = list(urlparse(url))
        value = parse_qs(fragments[4])
        value.update(params)
        fragments[4] = urlencode(value)
        url = urlunparse(fragments)
    return url


def to_setting_name(*names):
    return '_'.join([name.upper().replace('-', '_') for name in names if name])


def setting_name(*names):
    return to_setting_name(*((SETTING_PREFIX,) + names))


def sanitize_redirect(hosts, redirect_to):
    """
    Given a list of hostnames and an untrusted URL to redirect to,
    this method tests it to make sure it isn't garbage/harmful
    and returns it, else returns None, similar as how's it done
    on django.contrib.auth.views.
    """
    if redirect_to:
        try:
            # Don't redirect to a host that's not in the list
            netloc = urlparse(redirect_to)[1] or hosts[0]
        except (TypeError, AttributeError):
            pass
        else:
            if netloc in hosts:
                return redirect_to


def user_is_authenticated(user):
    if user and hasattr(user, 'is_authenticated'):
        if isinstance(user.is_authenticated, collections.Callable):
            authenticated = user.is_authenticated()
        else:
            authenticated = user.is_authenticated
    elif user:
        authenticated = True
    else:
        authenticated = False
    return authenticated


def user_is_active(user):
    if user and hasattr(user, 'is_active'):
        if isinstance(user.is_active, collections.Callable):
            is_active = user.is_active()
        else:
            is_active = user.is_active
    elif user:
        is_active = True
    else:
        is_active = False
    return is_active


# This slugify version was borrowed from django revision a61dbd6
def slugify(value):
    """Converts to lowercase, removes non-word characters (alphanumerics
    and underscores) and converts spaces to hyphens. Also strips leading
    and trailing whitespace."""
    value = unicodedata.normalize('NFKD', six.text_type(value)) \
                       .encode('ascii', 'ignore') \
                       .decode('ascii')
    value = re.sub('[^\w\s-]', '', value).strip().lower()
    return re.sub('[-\s]+', '-', value)


def first(func, items):
    """Return the first item in the list for what func returns True"""
    for item in items:
        if func(item):
            return item


def parse_qs(value):
    """Like urlparse.parse_qs but transform list values to single items"""
    return drop_lists(battery_parse_qs(value))


def drop_lists(value):
    out = {}
    for key, val in value.items():
        val = val[0]
        if isinstance(key, six.binary_type):
            key = six.text_type(key, 'utf-8')
        if isinstance(val, six.binary_type):
            val = six.text_type(val, 'utf-8')
        out[key] = val
    return out


def partial_pipeline_data(backend, user=None, partial_token=None,
                          *args, **kwargs):
    request_data = backend.strategy.request_data()

    partial_argument_name = backend.setting('PARTIAL_PIPELINE_TOKEN_NAME',
                                            'partial_token')
    partial_token = partial_token or \
        request_data.get(partial_argument_name) or \
        backend.strategy.session_get('partial_pipeline_token', None)

    if partial_token:
        partial = backend.strategy.partial_load(partial_token)
        partial_matches_request = False

        if partial and partial.backend == backend.name:
            partial_matches_request = True

            # Normally when resuming a pipeline, request_data will be empty. We
            # only need to check for a uid match if new data was provided (i.e.
            # if current request specifies the ID_KEY).
            if backend.ID_KEY in request_data:
                id_from_partial = partial.kwargs.get('uid')
                id_from_request = request_data.get(backend.ID_KEY)

                if id_from_partial != id_from_request:
                    partial_matches_request = False

        if partial_matches_request:
            if user:  # don't update user if it's None
                kwargs.setdefault('user', user)
            kwargs.setdefault('request', request_data)
            partial.extend_kwargs(kwargs)
            return partial
        else:
            backend.strategy.clean_partial_pipeline(partial_token)


def build_absolute_uri(host_url, path=None):
    """Build absolute URI with given (optional) path"""
    path = path or ''
    if path.startswith('http://') or path.startswith('https://'):
        return path
    if host_url.endswith('/') and path.startswith('/'):
        path = path[1:]
    return host_url + path


def constant_time_compare(val1, val2):
    """
    Returns True if the two strings are equal, False otherwise.
    The time taken is independent of the number of characters that match.
    This code was borrowed from Django 1.5.4-final
    """
    if len(val1) != len(val2):
        return False
    result = 0
    if six.PY3 and isinstance(val1, bytes) and isinstance(val2, bytes):
        for x, y in zip(val1, val2):
            result |= x ^ y
    else:
        for x, y in zip(val1, val2):
            result |= ord(x) ^ ord(y)
    return result == 0


def is_url(value):
    return value and \
           (value.startswith('http://') or
            value.startswith('https://') or
            value.startswith('/'))


def setting_url(backend, *names):
    for name in names:
        if is_url(name):
            return name
        else:
            value = backend.setting(name)
            if is_url(value):
                return value


def handle_http_errors(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except requests.HTTPError as err:
            if err.response.status_code == 400:
                raise AuthCanceled(args[0], response=err.response)
            elif err.response.status_code == 503:
                raise AuthUnreachableProvider(args[0])
            else:
                raise
    return wrapper


def append_slash(url):
    """Make sure we append a slash at the end of the URL otherwise we
    have issues with urljoin Example:
    >>> urlparse.urljoin('http://www.example.com/api/v3', 'user/1/')
    'http://www.example.com/api/user/1/'
    """
    if url and not url.endswith('/'):
        url = '{0}/'.format(url)
    return url


def get_strategy(strategy, storage, *args, **kwargs):
    Strategy = module_member(strategy)
    Storage = module_member(storage)
    return Strategy(Storage, *args, **kwargs)


class cache(object):
    """
    Cache decorator that caches the return value of a method for a
    specified time.

    It maintains a cache per class, so subclasses have a different cache entry
    for the same cached method.

    Does not work for methods with arguments.
    """
    def __init__(self, ttl):
        self.ttl = ttl
        self.cache = {}

    def __call__(self, fn):
        def wrapped(this):
            now = time.time()
            last_updated = None
            cached_value = None
            if this.__class__ in self.cache:
                last_updated, cached_value = self.cache[this.__class__]
            if not cached_value or now - last_updated > self.ttl:
                try:
                    cached_value = fn(this)
                    self.cache[this.__class__] = (now, cached_value)
                except:
                    # Use previously cached value when call fails, if available
                    if not cached_value:
                        raise
            return cached_value
        return wrapped



"""
Spotify backend, docs at:
    https://developer.spotify.com/spotify-web-api/
    https://developer.spotify.com/spotify-web-api/authorization-guide/
"""
import base64

from .oauth import BaseOAuth2


class SpotifyOAuth2(BaseOAuth2):
    """Spotify OAuth2 authentication backend"""
    name = 'spotify'
    ID_KEY = 'id'
    AUTHORIZATION_URL = 'https://accounts.spotify.com/authorize'
    ACCESS_TOKEN_URL = 'https://accounts.spotify.com/api/token'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ' '
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('refresh_token', 'refresh_token'),
    ]

    def auth_headers(self):
        auth_str = '{0}:{1}'.format(*self.get_key_and_secret())
        b64_auth_str = base64.urlsafe_b64encode(auth_str.encode()).decode()
        return {
            'Authorization': 'Basic {0}'.format(b64_auth_str)
        }

    def get_user_details(self, response):
        """Return user details from Spotify account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('display_name')
        )
        return {'username': response.get('id'),
                'email': response.get('email'),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://api.spotify.com/v1/me',
            headers={'Authorization': 'Bearer {0}'.format(access_token)}
        )



import json

from .oauth import OAuth2Test


class TaobaoOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.taobao.TAOBAOAuth'
    user_data_url = 'https://eco.taobao.com/router/rest'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'w2_expires_in': 0,
        'taobao_user_id': '1',
        'taobao_user_nick': 'foobar',
        'w1_expires_in': 1800,
        're_expires_in': 0,
        'r2_expires_in': 0,
        'expires_in': 86400,
        'r1_expires_in': 1800
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from .oauth import BaseOAuth1, BaseOAuth2


class VimeoOAuth1(BaseOAuth1):
    """Vimeo OAuth authentication backend"""
    name = 'vimeo'
    AUTHORIZATION_URL = 'https://vimeo.com/oauth/authorize'
    REQUEST_TOKEN_URL = 'https://vimeo.com/oauth/request_token'
    ACCESS_TOKEN_URL = 'https://vimeo.com/oauth/access_token'

    def get_user_id(self, details, response):
        return response.get('person', {}).get('id')

    def get_user_details(self, response):
        """Return user details from Twitter account"""
        person = response.get('person', {})
        fullname, first_name, last_name = self.get_user_names(
            person.get('display_name', '')
        )
        return {'username': person.get('username', ''),
                'email': '',
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        return self.get_json(
            'https://vimeo.com/api/rest/v2',
            params={'format': 'json', 'method': 'vimeo.people.getInfo'},
            auth=self.oauth_auth(access_token)
        )


class VimeoOAuth2(BaseOAuth2):
    """Vimeo OAuth2 authentication backend"""
    name = 'vimeo-oauth2'
    AUTHORIZATION_URL = 'https://api.vimeo.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://api.vimeo.com/oauth/access_token'
    REFRESH_TOKEN_URL = 'https://api.vimeo.com/oauth/request_token'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ','
    API_ACCEPT_HEADER = {'Accept': 'application/vnd.vimeo.*+json;version=3.0'}

    def get_redirect_uri(self, state=None):
        """
        Build redirect with redirect_state parameter.

        @Vimeo API 3 requires exact redirect uri without additional
        additional state parameter included
        """
        return self.redirect_uri

    def get_user_id(self, details, response):
        """Return user id"""
        try:
            user_id = response.get('user', {})['uri'].split('/')[-1]
        except KeyError:
            user_id = None
        return user_id

    def get_user_details(self, response):
        """Return user details from account"""
        user = response.get('user', {})
        fullname, first_name, last_name = self.get_user_names(
            user.get('name', '')
        )
        return {'username': fullname,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        return self.get_json(
            'https://api.vimeo.com/me',
            params={'access_token': access_token},
            headers=VimeoOAuth2.API_ACCEPT_HEADER,
        )



from .oauth import BaseOAuth2


class MineIDOAuth2(BaseOAuth2):
    """MineID OAuth2 authentication backend"""
    name = 'mineid'
    _AUTHORIZATION_URL = '%(scheme)s://%(host)s/oauth/authorize'
    _ACCESS_TOKEN_URL = '%(scheme)s://%(host)s/oauth/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ','
    EXTRA_DATA = [
    ]

    def get_user_details(self, response):
        """Return user details"""
        return {'email': response.get('email'),
                'username': response.get('email')}

    def user_data(self, access_token, *args, **kwargs):
        return self._user_data(access_token)

    def _user_data(self, access_token, path=None):
        url = '%(scheme)s://%(host)s/api/user' % self.get_mineid_url_params()
        return self.get_json(url, params={'access_token': access_token})

    @property
    def AUTHORIZATION_URL(self):
        return self._AUTHORIZATION_URL % self.get_mineid_url_params()

    @property
    def ACCESS_TOKEN_URL(self):
        return self._ACCESS_TOKEN_URL % self.get_mineid_url_params()

    def get_mineid_url_params(self):
        return {
            'host': self.setting('HOST', 'www.mineid.org'),
            'scheme': self.setting('SCHEME', 'https'),
        }



import datetime
import json
import unittest2

from httpretty import HTTPretty

from six.moves.urllib_parse import urlencode

from ...actions import do_disconnect

from ..models import User
from .oauth import OAuth1Test, OAuth2Test
from .open_id import OpenIdTest
from .open_id_connect import OpenIdConnectTestMixin, NO_JWKEST


class GoogleOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.google.GoogleOAuth2'
    user_data_url = 'https://www.googleapis.com/plus/v1/people/me'
    expected_username = 'foo'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'aboutMe': 'About me text',
        'cover': {
            'coverInfo': {
                'leftImageOffset': 0,
                'topImageOffset': 0
            },
            'coverPhoto': {
                'height': 629,
                'url': 'https://lh5.googleusercontent.com/-ui-GqpNh5Ms/'
                       'AAAAAAAAAAI/AAAAAAAAAZw/a7puhHMO_fg/photo.jpg',
                'width': 940
            },
            'layout': 'banner'
        },
        'displayName': 'Foo Bar',
        'emails': [{
            'type': 'account',
            'value': 'foo@bar.com'
        }],
        'etag': '"e-tag string"',
        'gender': 'male',
        'id': '101010101010101010101',
        'image': {
            'url': 'https://lh5.googleusercontent.com/-ui-GqpNh5Ms/'
                   'AAAAAAAAAAI/AAAAAAAAAZw/a7puhHMO_fg/photo.jpg',
        },
        'isPlusUser': True,
        'kind': 'plus#person',
        'language': 'en',
        'name': {
            'familyName': 'Bar',
            'givenName': 'Foo'
        },
        'objectType': 'person',
        'occupation': 'Software developer',
        'organizations': [{
            'name': 'Org name',
            'primary': True,
            'type': 'school'
        }],
        'placesLived': [{
            'primary': True,
            'value': 'Anyplace'
        }],
        'url': 'https://plus.google.com/101010101010101010101',
        'urls': [{
            'label': 'http://foobar.com',
            'type': 'otherProfile',
            'value': 'http://foobar.com',
        }],
        'verified': False
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_with_unique_user_id(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GOOGLE_OAUTH2_USE_UNIQUE_USER_ID': True,
        })
        self.do_login()


class GoogleOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.google.GoogleOAuth'
    user_data_url = 'https://www.googleapis.com/userinfo/email'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    user_data_body = urlencode({
        'email': 'foobar@gmail.com',
        'isVerified': 'true',
        'id': '101010101010101010101'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_with_unique_user_id(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GOOGLE_OAUTH_USE_UNIQUE_USER_ID': True
        })
        self.do_login()

    def test_with_anonymous_key_and_secret(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GOOGLE_OAUTH_KEY': None,
            'SOCIAL_AUTH_GOOGLE_OAUTH_SECRET': None
        })
        self.do_login()


JANRAIN_NONCE = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')


class GoogleOpenIdTest(OpenIdTest):
    backend_path = 'social_core.backends.google.GoogleOpenId'
    expected_username = 'FooBar'
    discovery_body = ''.join([
        '<?xml version="1.0" encoding="UTF-8"?>',
        '<xrds:XRDS xmlns:xrds="xri://$xrds" xmlns="xri://$xrd*($v*2.0)">',
        '<XRD>',
        '<Service priority="0">',
        '<Type>http://specs.openid.net/auth/2.0/signon</Type>',
        '<Type>http://openid.net/srv/ax/1.0</Type>',
        '<Type>http://specs.openid.net/extensions/ui/1.0/mode/popup</Type>',
        '<Type>http://specs.openid.net/extensions/ui/1.0/icon</Type>',
        '<Type>http://specs.openid.net/extensions/pape/1.0</Type>',
        '<URI>https://www.google.com/accounts/o8/ud</URI>',
        '</Service>',
        '<Service priority="10">',
        '<Type>http://specs.openid.net/auth/2.0/signon</Type>',
        '<Type>http://openid.net/srv/ax/1.0</Type>',
        '<Type>http://specs.openid.net/extensions/ui/1.0/mode/popup</Type>',
        '<Type>http://specs.openid.net/extensions/ui/1.0/icon</Type>',
        '<Type>http://specs.openid.net/extensions/pape/1.0</Type>',
        '<URI>https://www.google.com/accounts/o8/ud?source=mail</URI>',
        '</Service>',
        '<Service priority="10">',
        '<Type>http://specs.openid.net/auth/2.0/signon</Type>',
        '<Type>http://openid.net/srv/ax/1.0</Type>',
        '<Type>http://specs.openid.net/extensions/ui/1.0/mode/popup</Type>',
        '<Type>http://specs.openid.net/extensions/ui/1.0/icon</Type>',
        '<Type>http://specs.openid.net/extensions/pape/1.0</Type>',
        '<URI>https://www.google.com/accounts/o8/ud?source=gmail.com</URI>',
        '</Service>',
        '<Service priority="10">',
        '<Type>http://specs.openid.net/auth/2.0/signon</Type>',
        '<Type>http://openid.net/srv/ax/1.0</Type>',
        '<Type>http://specs.openid.net/extensions/ui/1.0/mode/popup</Type>',
        '<Type>http://specs.openid.net/extensions/ui/1.0/icon</Type>',
        '<Type>http://specs.openid.net/extensions/pape/1.0</Type>',
        '<URI>',
        'https://www.google.com/accounts/o8/ud?source=googlemail.com',
        '</URI>',
        '</Service>',
        '<Service priority="10">',
        '<Type>http://specs.openid.net/auth/2.0/signon</Type>',
        '<Type>http://openid.net/srv/ax/1.0</Type>',
        '<Type>http://specs.openid.net/extensions/ui/1.0/mode/popup</Type>',
        '<Type>http://specs.openid.net/extensions/ui/1.0/icon</Type>',
        '<Type>http://specs.openid.net/extensions/pape/1.0</Type>',
        '<URI>https://www.google.com/accounts/o8/ud?source=profiles</URI>',
        '</Service>',
        '</XRD>',
        '</xrds:XRDS>'
    ])
    server_response = urlencode({
        'janrain_nonce': JANRAIN_NONCE,
        'openid.assoc_handle': 'assoc-handle',
        'openid.claimed_id': 'https://www.google.com/accounts/o8/id?'
                             'id=some-google-id',
        'openid.ext1.mode': 'fetch_response',
        'openid.ext1.type.email': 'http://axschema.org/contact/email',
        'openid.ext1.type.first_name': 'http://axschema.org/namePerson/first',
        'openid.ext1.type.last_name': 'http://axschema.org/namePerson/last',
        'openid.ext1.type.old_email': 'http://schema.openid.net/contact/email',
        'openid.ext1.value.email': 'foo@bar.com',
        'openid.ext1.value.first_name': 'Foo',
        'openid.ext1.value.last_name': 'Bar',
        'openid.ext1.value.old_email': 'foo@bar.com',
        'openid.identity': 'https://www.google.com/accounts/o8/id?'
                           'id=some-google-id',
        'openid.mode': 'id_res',
        'openid.ns': 'http://specs.openid.net/auth/2.0',
        'openid.ns.ext1': 'http://openid.net/srv/ax/1.0',
        'openid.op_endpoint': 'https://www.google.com/accounts/o8/ud',
        'openid.response_nonce': JANRAIN_NONCE + 'by95cT34vX7p9g',
        'openid.return_to': 'http://myapp.com/complete/google/?'
                            'janrain_nonce=' + JANRAIN_NONCE,
        'openid.sig': 'brT2kmu3eCzb1gQ1pbaXdnWioVM=',
        'openid.signed': 'op_endpoint,claimed_id,identity,return_to,'
                         'response_nonce,assoc_handle,ns.ext1,ext1.mode,'
                         'ext1.type.old_email,ext1.value.old_email,'
                         'ext1.type.first_name,ext1.value.first_name,'
                         'ext1.type.last_name,ext1.value.last_name,'
                         'ext1.type.email,ext1.value.email'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class GoogleRevokeTokenTest(GoogleOAuth2Test):
    def test_revoke_token(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GOOGLE_OAUTH2_REVOKE_TOKENS_ON_DISCONNECT': True
        })
        self.do_login()
        user = User.get(self.expected_username)
        user.password = 'password'
        HTTPretty.register_uri(self._method(self.backend.REVOKE_TOKEN_METHOD),
                               self.backend.REVOKE_TOKEN_URL,
                               status=200)
        do_disconnect(self.backend, user)


@unittest2.skipIf(NO_JWKEST, 'No Jwkest installed')
class GoogleOpenIdConnectTest(OpenIdConnectTestMixin, GoogleOAuth2Test):
    backend_path = \
        'social_core.backends.google_openidconnect.GoogleOpenIdConnect'
    user_data_url = \
        'https://www.googleapis.com/plus/v1/people/me/openIdConnect'
    issuer = 'accounts.google.com'
    openid_config_body = ''.join([
        '{',
        ' "issuer": "https://accounts.google.com",',
        ' "authorization_endpoint": "https://accounts.google.com/o/oauth2/v2/auth",',
        ' "token_endpoint": "https://www.googleapis.com/oauth2/v4/token",',
        ' "userinfo_endpoint": "https://www.googleapis.com/oauth2/v3/userinfo",',
        ' "revocation_endpoint": "https://accounts.google.com/o/oauth2/revoke",',
        ' "jwks_uri": "https://www.googleapis.com/oauth2/v3/certs",',
        ' "response_types_supported": [',
        '  "code",',
        '  "token",',
        '  "id_token",',
        '  "code token",',
        '  "code id_token",',
        '  "token id_token",',
        '  "code token id_token",',
        '  "none"',
        ' ],',
        ' "subject_types_supported": [',
        '  "public"',
        ' ],',
        ' "id_token_signing_alg_values_supported": [',
        '  "RS256"',
        ' ],',
        ' "scopes_supported": [',
        '  "openid",',
        '  "email",',
        '  "profile"',
        ' ],',
        ' "token_endpoint_auth_methods_supported": [',
        '  "client_secret_post",',
        '  "client_secret_basic"',
        ' ],',
        ' "claims_supported": [',
        '  "aud",',
        '  "email",',
        '  "email_verified",',
        '  "exp",',
        '  "family_name",',
        '  "given_name",',
        '  "iat",',
        '  "iss",',
        '  "locale",',
        '  "name",',
        '  "picture",',
        '  "sub"',
        ' ]',
        '}'
    ])



import time
import random
import hashlib

from .utils import setting_name, module_member
from .store import OpenIdStore, OpenIdSessionWrapper
from .pipeline import DEFAULT_AUTH_PIPELINE, DEFAULT_DISCONNECT_PIPELINE
from .pipeline.utils import partial_load, partial_store, partial_prepare


class BaseTemplateStrategy(object):
    def __init__(self, strategy):
        self.strategy = strategy

    def render(self, tpl=None, html=None, context=None):
        if not tpl and not html:
            raise ValueError('Missing template or html parameters')
        context = context or {}
        if tpl:
            return self.render_template(tpl, context)
        else:
            return self.render_string(html, context)

    def render_template(self, tpl, context):
        raise NotImplementedError('Implement in subclass')

    def render_string(self, html, context):
        raise NotImplementedError('Implement in subclass')


class BaseStrategy(object):
    ALLOWED_CHARS = 'abcdefghijklmnopqrstuvwxyz' \
                    'ABCDEFGHIJKLMNOPQRSTUVWXYZ' \
                    '0123456789'
    DEFAULT_TEMPLATE_STRATEGY = BaseTemplateStrategy

    def __init__(self, storage=None, tpl=None):
        self.storage = storage
        self.tpl = (tpl or self.DEFAULT_TEMPLATE_STRATEGY)(self)

    def setting(self, name, default=None, backend=None):
        names = [setting_name(name), name]
        if backend:
            names.insert(0, setting_name(backend.name, name))
        for name in names:
            try:
                return self.get_setting(name)
            except (AttributeError, KeyError):
                pass
        return default

    def create_user(self, *args, **kwargs):
        return self.storage.user.create_user(*args, **kwargs)

    def get_user(self, *args, **kwargs):
        return self.storage.user.get_user(*args, **kwargs)

    def session_setdefault(self, name, value):
        self.session_set(name, value)
        return self.session_get(name)

    def openid_session_dict(self, name):
        # Many frameworks are switching the session serialization from Pickle
        # to JSON to avoid code execution risks. Flask did this from Flask
        # 0.10, Django is switching to JSON by default from version 1.6.
        #
        # Sadly python-openid stores classes instances in the session which
        # fails the JSON serialization, the classes are:
        #
        #   openid.yadis.manager.YadisServiceManager
        #   openid.consumer.discover.OpenIDServiceEndpoint
        #
        # This method will return a wrapper over the session value used with
        # openid (a dict) which will automatically keep a pickled value for the
        # mentioned classes.
        return OpenIdSessionWrapper(self.session_setdefault(name, {}))

    def to_session_value(self, val):
        return val

    def from_session_value(self, val):
        return val

    def partial_save(self, next_step, backend, *args, **kwargs):
        return partial_store(self, backend, next_step, *args, **kwargs)

    def partial_prepare(self, next_step, backend, *args, **kwargs):
        return partial_prepare(self, backend, next_step, *args, **kwargs)

    def partial_load(self, token):
        return partial_load(self, token)

    def clean_partial_pipeline(self, token):
        self.storage.partial.destroy(token)

    def openid_store(self):
        return OpenIdStore(self)

    def get_pipeline(self, backend=None):
        return self.setting('PIPELINE', DEFAULT_AUTH_PIPELINE, backend)

    def get_disconnect_pipeline(self, backend=None):
        return self.setting(
            'DISCONNECT_PIPELINE',
            DEFAULT_DISCONNECT_PIPELINE,
            backend
        )

    def random_string(self, length=12, chars=ALLOWED_CHARS):
        # Implementation borrowed from django 1.4
        try:
            random.SystemRandom()
        except NotImplementedError:
            key = self.setting('SECRET_KEY', '')
            seed = '{0}{1}{2}'.format(random.getstate(), time.time(), key)
            random.seed(hashlib.sha256(seed.encode()).digest())
        return ''.join([random.choice(chars) for i in range(length)])

    def absolute_uri(self, path=None):
        uri = self.build_absolute_uri(path)
        if uri and self.setting('REDIRECT_IS_HTTPS'):
            uri = uri.replace('http://', 'https://')
        return uri

    def get_language(self):
        """Return current language"""
        return ''

    def send_email_validation(self, backend, email, partial_token=None):
        email_validation = self.setting('EMAIL_VALIDATION_FUNCTION')
        send_email = module_member(email_validation)
        code = self.storage.code.make_code(email)
        send_email(self, backend, code, partial_token)
        return code

    def validate_email(self, email, code):
        verification_code = self.storage.code.get_code(code)
        if not verification_code or verification_code.code != code:
            return False
        elif verification_code.email != email:
            return False
        elif verification_code.verified:
            return False
        else:
            verification_code.verify()
            return True

    def render_html(self, tpl=None, html=None, context=None):
        """Render given template or raw html with given context"""
        return self.tpl.render(tpl, html, context)

    def authenticate(self, backend, *args, **kwargs):
        """Trigger the authentication mechanism tied to the current
        framework"""
        kwargs['strategy'] = self
        kwargs['storage'] = self.storage
        kwargs['backend'] = backend
        args, kwargs = self.clean_authenticate_args(*args, **kwargs)
        return backend.authenticate(*args, **kwargs)

    def clean_authenticate_args(self, *args, **kwargs):
        """Take authenticate arguments and return a "cleaned" version
        of them"""
        return args, kwargs

    def get_backends(self):
        """Return configured backends"""
        return self.setting('AUTHENTICATION_BACKENDS', [])

    # Implement the following methods on strategies sub-classes

    def redirect(self, url):
        """Return a response redirect to the given URL"""
        raise NotImplementedError('Implement in subclass')

    def get_setting(self, name):
        """Return value for given setting name"""
        raise NotImplementedError('Implement in subclass')

    def html(self, content):
        """Return HTTP response with given content"""
        raise NotImplementedError('Implement in subclass')

    def request_data(self, merge=True):
        """Return current request data (POST or GET)"""
        raise NotImplementedError('Implement in subclass')

    def request_host(self):
        """Return current host value"""
        raise NotImplementedError('Implement in subclass')

    def session_get(self, name, default=None):
        """Return session value for given key"""
        raise NotImplementedError('Implement in subclass')

    def session_set(self, name, value):
        """Set session value for given key"""
        raise NotImplementedError('Implement in subclass')

    def session_pop(self, name):
        """Pop session value for given key"""
        raise NotImplementedError('Implement in subclass')

    def build_absolute_uri(self, path=None):
        """Build absolute URI with given (optional) path"""
        raise NotImplementedError('Implement in subclass')

    def request_is_secure(self):
        """Is the request using HTTPS?"""
        raise NotImplementedError('Implement in subclass')

    def request_path(self):
        """path of the current request"""
        raise NotImplementedError('Implement in subclass')

    def request_port(self):
        """Port in use for this request"""
        raise NotImplementedError('Implement in subclass')

    def request_get(self):
        """Request GET data"""
        raise NotImplementedError('Implement in subclass')

    def request_post(self):
        """Request POST data"""
        raise NotImplementedError('Implement in subclass')



"""
Github OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/github.html
"""
from requests import HTTPError

from six.moves.urllib.parse import urljoin

from .oauth import BaseOAuth2
from ..exceptions import AuthFailed


class GithubOAuth2(BaseOAuth2):
    """Github OAuth authentication backend"""
    name = 'github'
    API_URL = 'https://api.github.com/'
    AUTHORIZATION_URL = 'https://github.com/login/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://github.com/login/oauth/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ','
    EXTRA_DATA = [
        ('id', 'id'),
        ('expires', 'expires'),
        ('login', 'login')
    ]

    def api_url(self):
        return self.API_URL

    def get_user_details(self, response):
        """Return user details from Github account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('name')
        )
        return {'username': response.get('login'),
                'email': response.get('email') or '',
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        data = self._user_data(access_token)
        if not data.get('email'):
            try:
                emails = self._user_data(access_token, '/emails')
            except (HTTPError, ValueError, TypeError):
                emails = []

            if emails:
                email = emails[0]
                primary_emails = [
                    e for e in emails
                    if not isinstance(e, dict) or e.get('primary')
                ]
                if primary_emails:
                    email = primary_emails[0]
                if isinstance(email, dict):
                    email = email.get('email', '')
                data['email'] = email
        return data

    def _user_data(self, access_token, path=None):
        url = urljoin(self.api_url(), 'user{0}'.format(path or ''))
        return self.get_json(url, params={'access_token': access_token})


class GithubMemberOAuth2(GithubOAuth2):
    no_member_string = ''

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        user_data = super(GithubMemberOAuth2, self).user_data(
            access_token, *args, **kwargs
        )
        try:
            self.request(self.member_url(user_data), params={
                'access_token': access_token
            })
        except HTTPError as err:
            # if the user is a member of the organization, response code
            # will be 204, see http://bit.ly/ZS6vFl
            if err.response.status_code != 204:
                raise AuthFailed(self,
                                 'User doesn\'t belong to the organization')
        return user_data

    def member_url(self, user_data):
        raise NotImplementedError('Implement in subclass')


class GithubOrganizationOAuth2(GithubMemberOAuth2):
    """Github OAuth2 authentication backend for organizations"""
    name = 'github-org'
    no_member_string = 'User doesn\'t belong to the organization'

    def member_url(self, user_data):
        return urljoin(
            self.api_url(),
            'orgs/{org}/members/{username}'.format(
                org=self.setting('NAME'),
                username=user_data.get('login')
            )
        )


class GithubTeamOAuth2(GithubMemberOAuth2):
    """Github OAuth2 authentication backend for teams"""
    name = 'github-team'
    no_member_string = 'User doesn\'t belong to the team'

    def member_url(self, user_data):
        return urljoin(
            self.api_url(),
            'teams/{team_id}/members/{username}'.format(
                team_id=self.setting('ID'),
                username=user_data.get('login')
            )
        )



"""
500px OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/five_hundred_px.html
"""
from .oauth import BaseOAuth1


class FiveHundredPxOAuth(BaseOAuth1):
    """500px OAuth authentication backend"""
    name = '500px'
    AUTHORIZATION_URL = 'https://api.500px.com/v1/oauth/authorize'
    REQUEST_TOKEN_URL = 'https://api.500px.com/v1/oauth/request_token'
    ACCESS_TOKEN_URL = 'https://api.500px.com/v1/oauth/access_token'

    def get_user_details(self, user):
        """Return user details from 500px account"""
        fullname, first_name, last_name = self.get_user_names(
            user.get('fullname')
        )
        return {
            'username': user.get('username') or user.get('id'),
            'email': user.get('email'),
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        response = self.get_json(
            'https://api.500px.com/v1/users',
            auth=self.oauth_auth(access_token)
        )
        return response.get('user')



from ..pipeline.partial import partial


@partial
def ask_for_password(strategy, *args, **kwargs):
    if strategy.session_get('password'):
        return {'password': strategy.session_get('password')}
    else:
        return strategy.redirect(strategy.build_absolute_uri('/password'))


@partial
def ask_for_slug(strategy, *args, **kwargs):
    if strategy.session_get('slug'):
        return {'slug': strategy.session_get('slug')}
    else:
        return strategy.redirect(strategy.build_absolute_uri('/slug'))


def set_password(strategy, user, *args, **kwargs):
    user.set_password(kwargs['password'])


def set_slug(strategy, user, *args, **kwargs):
    user.slug = kwargs['slug']


def remove_user(strategy, user, *args, **kwargs):
    return {'user': None}


@partial
def set_user_from_kwargs(strategy, *args, **kwargs):
    if strategy.session_get('attribute'):
        kwargs['user'].id
    else:
        return strategy.redirect(strategy.build_absolute_uri('/attribute'))


@partial
def set_user_from_args(strategy, user, *args, **kwargs):
    if strategy.session_get('attribute'):
        user.id
    else:
        return strategy.redirect(strategy.build_absolute_uri('/attribute'))



import sys
import unittest2 as unittest

try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock

from ..utils import sanitize_redirect, user_is_authenticated, \
                    user_is_active, slugify, build_absolute_uri, \
                    partial_pipeline_data
from .models import TestPartial


PY3 = sys.version_info[0] == 3


class SanitizeRedirectTest(unittest.TestCase):
    def test_none_redirect(self):
        self.assertEqual(sanitize_redirect(['myapp.com'], None), None)

    def test_empty_redirect(self):
        self.assertEqual(sanitize_redirect(['myapp.com'], ''), None)

    def test_dict_redirect(self):
        self.assertEqual(sanitize_redirect(['myapp.com'], {}), None)

    def test_invalid_redirect(self):
        self.assertEqual(sanitize_redirect(['myapp.com'], {'foo': 'bar'}), None)

    def test_wrong_path_redirect(self):
        self.assertEqual(
            sanitize_redirect(['myapp.com'], 'http://notmyapp.com/path/'),
            None
        )

    def test_valid_absolute_redirect(self):
        self.assertEqual(
            sanitize_redirect(['myapp.com'], 'http://myapp.com/path/'),
            'http://myapp.com/path/'
        )

    def test_valid_relative_redirect(self):
        self.assertEqual(sanitize_redirect(['myapp.com'], '/path/'), '/path/')

    def test_multiple_hosts(self):
        allowed_hosts = ['myapp1.com', 'myapp2.com']
        for host in allowed_hosts:
            url = 'http://{}/path/'.format(host)
            self.assertEqual(sanitize_redirect(allowed_hosts, url), url)

    def test_multiple_hosts_wrong_host(self):
        self.assertEqual(sanitize_redirect(
            ['myapp1.com', 'myapp2.com'], 'http://notmyapp.com/path/'), None)


class UserIsAuthenticatedTest(unittest.TestCase):
    def test_user_is_none(self):
        self.assertEqual(user_is_authenticated(None), False)

    def test_user_is_not_none(self):
        self.assertEqual(user_is_authenticated(object()), True)

    def test_user_has_is_authenticated(self):
        class User(object):
            is_authenticated = True
        self.assertEqual(user_is_authenticated(User()), True)

    def test_user_has_is_authenticated_callable(self):
        class User(object):
            def is_authenticated(self):
                return True
        self.assertEqual(user_is_authenticated(User()), True)


class UserIsActiveTest(unittest.TestCase):
    def test_user_is_none(self):
        self.assertEqual(user_is_active(None), False)

    def test_user_is_not_none(self):
        self.assertEqual(user_is_active(object()), True)

    def test_user_has_is_active(self):
        class User(object):
            is_active = True
        self.assertEqual(user_is_active(User()), True)

    def test_user_has_is_active_callable(self):
        class User(object):
            def is_active(self):
                return True
        self.assertEqual(user_is_active(User()), True)


class SlugifyTest(unittest.TestCase):
    def test_slugify_formats(self):
        if PY3:
            self.assertEqual(slugify('FooBar'), 'foobar')
            self.assertEqual(slugify('Foo Bar'), 'foo-bar')
            self.assertEqual(slugify('Foo (Bar)'), 'foo-bar')
        else:
            self.assertEqual(slugify('FooBar'.decode('utf-8')), 'foobar')
            self.assertEqual(slugify('Foo Bar'.decode('utf-8')), 'foo-bar')
            self.assertEqual(slugify('Foo (Bar)'.decode('utf-8')), 'foo-bar')


class BuildAbsoluteURITest(unittest.TestCase):
    def setUp(self):
        self.host = 'http://foobar.com'

    def tearDown(self):
        self.host = None

    def test_path_none(self):
        self.assertEqual(build_absolute_uri(self.host), self.host)

    def test_path_empty(self):
        self.assertEqual(build_absolute_uri(self.host, ''), self.host)

    def test_path_http(self):
        self.assertEqual(build_absolute_uri(self.host, 'http://barfoo.com'),
                         'http://barfoo.com')

    def test_path_https(self):
        self.assertEqual(build_absolute_uri(self.host, 'https://barfoo.com'),
                         'https://barfoo.com')

    def test_host_ends_with_slash_and_path_starts_with_slash(self):
        self.assertEqual(build_absolute_uri(self.host + '/', '/foo/bar'),
                         'http://foobar.com/foo/bar')

    def test_absolute_uri(self):
        self.assertEqual(build_absolute_uri(self.host, '/foo/bar'),
                         'http://foobar.com/foo/bar')


class PartialPipelineData(unittest.TestCase):
    def test_returns_partial_when_uid_and_email_do_match(self):
        email = 'foo@example.com'
        backend = self._backend({'uid': email})
        backend.strategy.request_data.return_value = {
            backend.ID_KEY: email
        }
        key, val = ('foo', 'bar')
        partial = partial_pipeline_data(backend, None,
                                        *(), **dict([(key, val)]))
        self.assertTrue(key in partial.kwargs)
        self.assertEqual(partial.kwargs[key], val)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 0)

    def test_clean_pipeline_when_uid_does_not_match(self):
        backend = self._backend({'uid': 'foo@example.com'})
        backend.strategy.request_data.return_value = {
            backend.ID_KEY: 'bar@example.com'
        }
        key, val = ('foo', 'bar')
        partial = partial_pipeline_data(backend, None,
                                        *(), **dict([(key, val)]))
        self.assertIsNone(partial)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 1)

    def test_kwargs_included_in_result(self):
        backend = self._backend()
        key, val = ('foo', 'bar')
        partial = partial_pipeline_data(backend, None,
                                        *(), **dict([(key, val)]))
        self.assertTrue(key in partial.kwargs)
        self.assertEqual(partial.kwargs[key], val)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 0)

    def test_update_user(self):
        user = object()
        backend = self._backend(session_kwargs={'user': None})
        partial = partial_pipeline_data(backend, user)
        self.assertTrue('user' in partial.kwargs)
        self.assertEqual(partial.kwargs['user'], user)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 0)

    def _backend(self, session_kwargs=None):
        backend = Mock()
        backend.ID_KEY = 'email'
        backend.name = 'mock-backend'

        strategy = Mock()
        strategy.request = None
        strategy.request_data.return_value = {}
        strategy.session_get.return_value = object()
        strategy.partial_load.return_value = TestPartial.prepare(backend.name, 0, {
            'args': [],
            'kwargs': session_kwargs or {}
        })

        backend.strategy = strategy
        return backend



"""
Slack OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/slack.html
    https://api.slack.com/docs/oauth
"""
from .oauth import BaseOAuth2


class SlackOAuth2(BaseOAuth2):
    """Slack OAuth authentication backend"""
    name = 'slack'
    AUTHORIZATION_URL = 'https://slack.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://slack.com/api/oauth.access'
    ACCESS_TOKEN_METHOD = 'POST'
    DEFAULT_SCOPE = ['identity.basic', 'identity.email']
    SCOPE_SEPARATOR = ','
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('id', 'id'),
        ('name', 'name'),
        ('real_name', 'real_name')
    ]

    def auth_extra_arguments(self):
        params = super(SlackOAuth2, self).auth_extra_arguments() or {}
        if self.setting('TEAM'):
            params['team'] = self.setting('TEAM')
        return params

    def get_user_details(self, response):
        """Return user details from Slack account"""
        # Build the username with the team $username@$team_url
        # Necessary to get unique names for all of slack
        user = response['user']
        team = response.get('team')
        name = user['name']
        email = user.get('email')
        username = email and email.split('@', 1)[0] or name
        fullname, first_name, last_name = self.get_user_names(name)

        if self.setting('USERNAME_WITH_TEAM', True) and team and \
           'name' in team:
            name = '{0}@{1}'.format(name, response['team']['name'])

        return {
            'username': username,
            'email': email,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        response = self.get_json('https://slack.com/api/users.identity',
                                 params={'token': access_token})
        if not response.get('id', None):
            response['id'] = response['user']['id']
        return response



# -*- coding: utf-8 -*-
"""
Professionaly OAuth 2.0 support.

This contribution adds support for professionaly.ru OAuth 2.0.
Username is retrieved from the identity returned by server.
"""
from time import time

from ..utils import parse_qs
from .oauth import BaseOAuth2


class ProfessionaliOAuth2(BaseOAuth2):
    name = 'professionali'
    ID_KEY = 'user_id'
    AUTHORIZATION_URL = 'https://api.professionali.ru/oauth/authorize.html'
    ACCESS_TOKEN_URL = 'https://api.professionali.ru/oauth/getToken.json'
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [
        ('avatar_big', 'avatar_big'),
        ('link', 'link')
    ]

    def get_user_details(self, response):
        first_name, last_name = map(response.get, ('firstname', 'lastname'))
        email = ''
        if self.setting('FAKE_EMAIL'):
            email = '{0}@professionali.ru'.format(time())
        return {
            'username': '{0}_{1}'.format(last_name, first_name),
            'first_name': first_name,
            'last_name': last_name,
            'email': email
        }

    def user_data(self, access_token, response, *args, **kwargs):
        url = 'https://api.professionali.ru/v6/users/get.json'
        fields = list(set(['firstname', 'lastname', 'avatar_big', 'link'] +
                          self.setting('EXTRA_DATA', [])))
        params = {
            'fields': ','.join(fields),
            'access_token': access_token,
            'ids[]': response['user_id']
        }
        try:
            return self.get_json(url, params)[0]
        except (TypeError, KeyError, IOError, ValueError, IndexError):
            return None

    def get_json(self, url, *args, **kwargs):
        return self.request(url, verify=False, *args, **kwargs).json()

    def get_querystring(self, url, *args, **kwargs):
        return parse_qs(self.request(url, verify=False, *args, **kwargs).text)



from .oauth import BaseOAuth2


class EchosignOAuth2(BaseOAuth2):
    name = 'echosign'
    REDIRECT_STATE = False
    ACCESS_TOKEN_METHOD = 'POST'
    REFRESH_TOKEN_METHOD = 'POST'
    REVOKE_TOKEN_METHOD = 'POST'
    AUTHORIZATION_URL = 'https://secure.echosign.com/public/oauth'
    ACCESS_TOKEN_URL = 'https://secure.echosign.com/oauth/token'
    REFRESH_TOKEN_URL = 'https://secure.echosign.com/oauth/refresh'
    REVOKE_TOKEN_URL = 'https://secure.echosign.com/oauth/revoke'

    def get_user_details(self, response):
        return response

    def get_user_id(self, details, response):
        return details['userInfoList'][0]['userId']

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            'https://api.echosign.com/api/rest/v3/users',
            headers={'Access-Token': access_token})



import json

from .oauth import OAuth2Test


class NaverOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.naver.NaverOAuth2'
    user_data_url = 'https://openapi.naver.com/v1/nid/getUserProfile.xml'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
    })

    user_data_content_type = 'text/xml'
    user_data_body = \
    '<?xml version="1.0" encoding="UTF-8" ?>' \
    '<data>' \
        '<result>' \
            '<resultcode>00</resultcode>' \
            '<message>success</message>' \
        '</result>' \
        '<response>' \
            '<email><![CDATA[foobar@naver.com]]></email>' \
            '<nickname><![CDATA[foobar]]></nickname>' \
            '<profile_image>' \
                '<![CDATA[http://naver.com/image.url.jpg]]>' \
            '</profile_image>' \
            '<age><![CDATA[20-29]]></age>' \
            '<gender>M</gender>' \
            '<id><![CDATA[123456]]></id>' \
            '<name><![CDATA[foobar]]></name>' \
            '<birthday><![CDATA[12-01]]></birthday>' \
        '</response>' \
    '</data>'

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class DribbbleOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.dribbble.DribbbleOAuth2'
    user_data_url = 'https://api.dribbble.com/v1/user'
    expected_username = 'foobar'

    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })

    user_data_body = json.dumps({
        'id': 'foobar',
        'username': 'foobar',
        'name': 'Foo Bar'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import datetime

from .oauth import BaseOAuth2


class AsanaOAuth2(BaseOAuth2):
    name = 'asana'
    AUTHORIZATION_URL = 'https://app.asana.com/-/oauth_authorize'
    ACCESS_TOKEN_METHOD = 'POST'
    ACCESS_TOKEN_URL = 'https://app.asana.com/-/oauth_token'
    REFRESH_TOKEN_URL = 'https://app.asana.com/-/oauth_token'
    REDIRECT_STATE = False
    USER_DATA_URL = 'https://app.asana.com/api/1.0/users/me'
    EXTRA_DATA = [
        ('expires_in', 'expires'),
        ('refresh_token', 'refresh_token'),
        ('name', 'name'),
    ]

    def get_user_details(self, response):
        data = response['data']
        fullname, first_name, last_name = self.get_user_names(data['name'])
        return {'email': data['email'],
                'username': data['email'],
                'fullname': fullname,
                'last_name': last_name,
                'first_name': first_name}

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(self.USER_DATA_URL, headers={
            'Authorization': 'Bearer {}'.format(access_token)
        })

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        data = super(AsanaOAuth2, self).extra_data(
            user, uid, response, details
        )
        if self.setting('ESTIMATE_EXPIRES_ON'):
            expires_on = datetime.datetime.utcnow() + \
                         datetime.timedelta(seconds=data['expires'])
            data['expires_on'] = expires_on.isoformat()
        return data



"""
Khan Academy OAuth backend, docs at:
    https://github.com/Khan/khan-api/wiki/Khan-Academy-API-Authentication
"""
import six

from six.moves.urllib_parse import urlencode

from oauthlib.oauth1 import SIGNATURE_HMAC, SIGNATURE_TYPE_QUERY
from requests_oauthlib import OAuth1

from .oauth import BaseOAuth1


class BrowserBasedOAuth1(BaseOAuth1):
    """Browser based mechanism OAuth authentication, fill the needed
    parameters to communicate properly with authentication service.

        REQUEST_TOKEN_URL       Request token URL (opened in web browser)
        ACCESS_TOKEN_URL        Access token URL
    """
    REQUEST_TOKEN_URL = ''
    OAUTH_TOKEN_PARAMETER_NAME = 'oauth_token'
    REDIRECT_URI_PARAMETER_NAME = 'redirect_uri'
    ACCESS_TOKEN_URL = ''

    def auth_url(self):
        """Return redirect url"""
        return self.unauthorized_token_request()

    def get_unauthorized_token(self):
        return self.strategy.request_data()

    def unauthorized_token_request(self):
        """Return request for unauthorized token (first stage)"""

        params = self.request_token_extra_arguments()
        params.update(self.get_scope_argument())
        key, secret = self.get_key_and_secret()
        # decoding='utf-8' produces errors with python-requests on Python3
        # since the final URL will be of type bytes
        decoding = None if six.PY3 else 'utf-8'
        state = self.get_or_create_state()
        auth = OAuth1(
            key,
            secret,
            callback_uri=self.get_redirect_uri(state),
            decoding=decoding,
            signature_method=SIGNATURE_HMAC,
            signature_type=SIGNATURE_TYPE_QUERY
        )
        url = self.REQUEST_TOKEN_URL + '?' + urlencode(params)
        url, _, _ = auth.client.sign(url)
        return url

    def oauth_auth(self, token=None, oauth_verifier=None):
        key, secret = self.get_key_and_secret()
        oauth_verifier = oauth_verifier or self.data.get('oauth_verifier')
        token = token or {}
        # decoding='utf-8' produces errors with python-requests on Python3
        # since the final URL will be of type bytes
        decoding = None if six.PY3 else 'utf-8'
        state = self.get_or_create_state()
        return OAuth1(key, secret,
                      resource_owner_key=token.get('oauth_token'),
                      resource_owner_secret=token.get('oauth_token_secret'),
                      callback_uri=self.get_redirect_uri(state),
                      verifier=oauth_verifier,
                      signature_method=SIGNATURE_HMAC,
                      signature_type=SIGNATURE_TYPE_QUERY,
                      decoding=decoding)


class KhanAcademyOAuth1(BrowserBasedOAuth1):
    """
    Class used for autorising with Khan Academy.

    Flow of Khan Academy is a bit different than most OAuth 1.0 and consinsts
    of the following steps:
    1. Create signed params to attach to the REQUEST_TOKEN_URL
    2. Redirect user to the REQUEST_TOKEN_URL that will respond with
       oauth_secret, oauth_token, oauth_verifier that should be used with
       ACCESS_TOKEN_URL
    3. Go to ACCESS_TOKEN_URL and grab oauth_token_secret.

    Note that we don't use the AUTHORIZATION_URL.

    REQUEST_TOKEN_URL requires the following arguments:
    oauth_consumer_key - Your app's consumer key
    oauth_nonce - Random 64-bit, unsigned number encoded as an ASCII string
        in decimal format. The nonce/timestamp pair should always be unique.
    oauth_version - OAuth version used by your app. Must be "1.0" for now.
    oauth_signature - String generated using the referenced signature method.
    oauth_signature_method - Signature algorithm (currently only support
        "HMAC-SHA1")
    oauth_timestamp - Integer representing the time the request is sent.
        The timestamp should be expressed in number of seconds
        after January 1, 1970 00:00:00 GMT.
    oauth_callback (optional) - URL to redirect to after request token is
        received and authorized by the user's chosen identity provider.
    """
    name = 'khanacademy-oauth1'
    ID_KEY = 'user_id'
    REQUEST_TOKEN_URL = 'http://www.khanacademy.org/api/auth/request_token'
    ACCESS_TOKEN_URL = 'https://www.khanacademy.org/api/auth/access_token'
    REDIRECT_URI_PARAMETER_NAME = 'oauth_callback'
    USER_DATA_URL = 'https://www.khanacademy.org/api/v1/user'

    EXTRA_DATA = [('user_id', 'user_id')]

    def get_user_details(self, response):
        """Return user details from Khan Academy account"""
        return {
            'username': response.get('email'),
            'email': response.get('email'),
            'fullname': response.get('nickname'),
            'first_name': '',
            'last_name': '',
            'user_id': response.get('user_id')
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        auth = self.oauth_auth(access_token)
        url, _, _ = auth.client.sign(self.USER_DATA_URL)
        return self.get_json(url)



"""
EVE Online Single Sign-On (SSO) OAuth2 backend
Documentation at https://eveonline-third-party-documentation.readthedocs.io/en/latest/sso/index.html
"""
from .oauth import BaseOAuth2


class EVEOnlineOAuth2(BaseOAuth2):
    """EVE Online OAuth authentication backend"""
    name = 'eveonline'
    BASE_URL = 'https://login.eveonline.com/oauth'
    AUTHORIZATION_URL = BASE_URL + '/authorize'
    ACCESS_TOKEN_URL = BASE_URL + '/token'
    ID_KEY = 'CharacterID'
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [
        ('CharacterID', 'id'),
        ('ExpiresOn', 'expires'),
        ('CharacterOwnerHash', 'owner_hash', True),
        ('refresh_token', 'refresh_token', True),
    ]

    def get_user_details(self, response):
        """Return user details from EVE Online account"""
        user_data = self.user_data(response['access_token'])
        fullname, first_name, last_name = self.get_user_names(
            user_data['CharacterName']
        )
        return {
            'email': '',
            'username': fullname,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def user_data(self, access_token, *args, **kwargs):
        """Get Character data from EVE server"""
        return self.get_json(
            'https://login.eveonline.com/oauth/verify',
            headers={'Authorization': 'Bearer {0}'.format(access_token)}
        )



from social.backends.oauth import BaseOAuth2


class MonzoOAuth2(BaseOAuth2):
    """
    Monzo OAuth2 authentication backend.
    """

    name = 'monzo'

    AUTHORIZATION_URL = 'https://auth.getmondo.co.uk/'
    ACCESS_TOKEN_URL = 'https://api.monzo.com/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response['accounts'][0]['description'],
        )

        return {
            'username': str(response.get('user_id')),
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            'https://api.monzo.com/accounts',
            headers={'Authorization': 'Bearer {0}'.format(access_token)},
        )



"""
Instagram OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/instagram.html
"""
import hmac

from hashlib import sha256

from .oauth import BaseOAuth2


class InstagramOAuth2(BaseOAuth2):
    name = 'instagram'
    AUTHORIZATION_URL = 'https://api.instagram.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://api.instagram.com/oauth/access_token'
    ACCESS_TOKEN_METHOD = 'POST'

    def get_user_id(self, details, response):
        # Sometimes Instagram returns 'user', sometimes 'data', but API docs
        # says 'data' http://instagram.com/developer/endpoints/users/#get_users
        user = response.get('user') or response.get('data') or {}
        return user.get('id')

    def get_user_details(self, response):
        """Return user details from Instagram account"""
        # Sometimes Instagram returns 'user', sometimes 'data', but API docs
        # says 'data' http://instagram.com/developer/endpoints/users/#get_users
        user = response.get('user') or response.get('data') or {}
        username = user['username']
        email = user.get('email', '')
        fullname, first_name, last_name = self.get_user_names(
            user.get('full_name', '')
        )
        return {'username': username,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': email}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        key, secret = self.get_key_and_secret()
        params = {'access_token': access_token}
        sig = self._generate_sig("/users/self", params, secret)
        params['sig'] = sig
        return self.get_json('https://api.instagram.com/v1/users/self',
                             params=params)

    def _generate_sig(self, endpoint, params, secret):
        sig = endpoint
        for key in sorted(params.keys()):
            sig += '|%s=%s' % (key, params[key])
        return hmac.new(secret.encode(), sig.encode(), sha256).hexdigest()



import datetime
from calendar import timegm

import six

from jwkest import JWKESTException
from jwkest.jwk import KEYS
from jwkest.jws import JWS

from .oauth import BaseOAuth2
from ..utils import cache
from ..exceptions import AuthTokenError


class OpenIdConnectAssociation(object):
    """ Use Association model to save the nonce by force."""
    def __init__(self, handle, secret='', issued=0, lifetime=0, assoc_type=''):
        self.handle = handle  # as nonce
        self.secret = secret.encode()  # not use
        self.issued = issued  # not use
        self.lifetime = lifetime  # not use
        self.assoc_type = assoc_type  # as state


class OpenIdConnectAuth(BaseOAuth2):
    """
    Base class for Open ID Connect backends.
    Currently only the code response type is supported.
    """
    # Override OIDC_ENDPOINT in your subclass to enable autoconfig of OIDC
    OIDC_ENDPOINT = None
    ID_TOKEN_MAX_AGE = 600
    DEFAULT_SCOPE = ['openid', 'profile', 'email']
    EXTRA_DATA = ['id_token', 'refresh_token', ('sub', 'id')]
    REDIRECT_STATE = False
    ACCESS_TOKEN_METHOD = 'POST'
    REVOKE_TOKEN_METHOD = 'GET'
    ID_KEY = 'sub'
    USERNAME_KEY = 'preferred_username'
    ID_TOKEN_ISSUER = ''
    ACCESS_TOKEN_URL = ''
    AUTHORIZATION_URL = ''
    REVOKE_TOKEN_URL = ''
    USERINFO_URL = ''
    JWKS_URI = ''

    def __init__(self, *args, **kwargs):
        self.id_token = None
        super(OpenIdConnectAuth, self).__init__(*args, **kwargs)

    def authorization_url(self):
        return self.AUTHORIZATION_URL or \
            self.oidc_config().get('authorization_endpoint')

    def access_token_url(self):
        return self.ACCESS_TOKEN_URL or \
            self.oidc_config().get('token_endpoint')

    def revoke_token_url(self, token, uid):
        return self.REVOKE_TOKEN_URL or \
            self.oidc_config().get('revocation_endpoint')

    def id_token_issuer(self):
        return self.ID_TOKEN_ISSUER or \
            self.oidc_config().get('issuer')

    def userinfo_url(self):
        return self.USERINFO_URL or \
            self.oidc_config().get('userinfo_endpoint')

    def jwks_uri(self):
        return self.JWKS_URI or \
            self.oidc_config().get('jwks_uri')

    @cache(ttl=86400)
    def oidc_config(self):
        return self.get_json(self.OIDC_ENDPOINT +
                             '/.well-known/openid-configuration')

    @cache(ttl=86400)
    def get_jwks_keys(self):
        keys = KEYS()
        keys.load_from_url(self.jwks_uri())

        # Add client secret as oct key so it can be used for HMAC signatures
        client_id, client_secret = self.get_key_and_secret()
        keys.add({'key': client_secret, 'kty': 'oct'})
        return keys

    def auth_params(self, state=None):
        """Return extra arguments needed on auth process."""
        params = super(OpenIdConnectAuth, self).auth_params(state)
        params['nonce'] = self.get_and_store_nonce(
            self.authorization_url(), state
        )
        return params

    def get_and_store_nonce(self, url, state):
        # Create a nonce
        nonce = self.strategy.random_string(64)
        # Store the nonce
        association = OpenIdConnectAssociation(nonce, assoc_type=state)
        self.strategy.storage.association.store(url, association)
        return nonce

    def get_nonce(self, nonce):
        try:
            return self.strategy.storage.association.get(
                server_url=self.authorization_url(),
                handle=nonce
            )[0]
        except IndexError:
            pass

    def remove_nonce(self, nonce_id):
        self.strategy.storage.association.remove([nonce_id])

    def validate_claims(self, id_token):
        if id_token['iss'] != self.id_token_issuer():
            raise AuthTokenError(self, 'Invalid issuer')

        client_id, client_secret = self.get_key_and_secret()

        if isinstance(id_token['aud'], six.string_types):
            id_token['aud'] = [id_token['aud']]

        if client_id not in id_token['aud']:
            raise AuthTokenError(self, 'Invalid audience')

        if len(id_token['aud']) > 1 and 'azp' not in id_token:
            raise AuthTokenError(self, 'Incorrect id_token: azp')

        if 'azp' in id_token and id_token['azp'] != client_id:
            raise AuthTokenError(self, 'Incorrect id_token: azp')

        utc_timestamp = timegm(datetime.datetime.utcnow().utctimetuple())
        if utc_timestamp > id_token['exp']:
            raise AuthTokenError(self, 'Signature has expired')

        if 'nbf' in id_token and utc_timestamp < id_token['nbf']:
            raise AuthTokenError(self, 'Incorrect id_token: nbf')

        # Verify the token was issued in the last 10 minutes
        iat_leeway = self.setting('ID_TOKEN_MAX_AGE', self.ID_TOKEN_MAX_AGE)
        if utc_timestamp > id_token['iat'] + iat_leeway:
            raise AuthTokenError(self, 'Incorrect id_token: iat')

        # Validate the nonce to ensure the request was not modified
        nonce = id_token.get('nonce')
        if not nonce:
            raise AuthTokenError(self, 'Incorrect id_token: nonce')

        nonce_obj = self.get_nonce(nonce)
        if nonce_obj:
            self.remove_nonce(nonce_obj.id)
        else:
            raise AuthTokenError(self, 'Incorrect id_token: nonce')

    def validate_and_return_id_token(self, jws):
        """
        Validates the id_token according to the steps at
        http://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation.
        """
        try:
            # Decode the JWT and raise an error if the sig is invalid
            id_token = JWS().verify_compact(jws.encode('utf-8'),
                                            self.get_jwks_keys())
        except JWKESTException:
            raise AuthTokenError(self, 'Signature verification failed')

        self.validate_claims(id_token)

        return id_token

    def request_access_token(self, *args, **kwargs):
        """
        Retrieve the access token. Also, validate the id_token and
        store it (temporarily).
        """
        response = self.get_json(*args, **kwargs)
        self.id_token = self.validate_and_return_id_token(response['id_token'])
        return response

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(self.userinfo_url(), headers={
            'Authorization': 'Bearer {0}'.format(access_token)
        })

    def get_user_details(self, response):
        username_key = self.setting('USERNAME_KEY', default=self.USERNAME_KEY)
        return {
            'username': response.get(username_key),
            'email': response.get('email'),
            'fullname': response.get('name'),
            'first_name': response.get('given_name'),
            'last_name': response.get('family_name'),
        }



import json

from .oauth import OAuth2Test


class CoinbaseOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.coinbase.CoinbaseOAuth2'
    user_data_url = 'https://coinbase.com/api/v1/users'
    expected_username = 'SatoshiNakamoto'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'users': [
            {
                'user': {
                    'id': "1A1zP1eP5QGefi2DMPTfTL5SLmv7DivfNa",
                    'name': "Satoshi Nakamoto",
                    'email': "satoshi@nakamoto.com",
                    'pin': None,
                    'time_zone': "Eastern Time (US & Canada)",
                    'native_currency': "USD",
                    'buy_level': 2,
                    'sell_level': 2,
                    'balance': {
                        'amount': "1000000",
                        'currency': "BTC"
                    },
                    'buy_limit': {
                        'amount': "50.00000000",
                        'currency': "BTC"
                    },
                    'sell_limit': {
                        'amount': "50.00000000",
                        'currency': "BTC"
                    }
                }
            }
        ]
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Skyrock OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/skyrock.html
"""
from .oauth import BaseOAuth1


class SkyrockOAuth(BaseOAuth1):
    """Skyrock OAuth authentication backend"""
    name = 'skyrock'
    ID_KEY = 'id_user'
    AUTHORIZATION_URL = 'https://api.skyrock.com/v2/oauth/authenticate'
    REQUEST_TOKEN_URL = 'https://api.skyrock.com/v2/oauth/initiate'
    ACCESS_TOKEN_URL = 'https://api.skyrock.com/v2/oauth/token'
    EXTRA_DATA = [('id', 'id')]

    def get_user_details(self, response):
        """Return user details from Skyrock account"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response['firstname'],
            last_name=response['name']
        )
        return {'username': response['username'],
                'email': response['email'],
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token):
        """Return user data provided"""
        return self.get_json('https://api.skyrock.com/v2/user/get.json',
                             auth=self.oauth_auth(access_token))
